package com.qbizm.kramerius.webclient.repository;

import java.io.InputStream;
import java.math.BigInteger;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.axis.types.NonNegativeInteger;
import org.apache.log4j.Logger;

import com.kramerius.webclient.fedoraws.ArrayOfString;
import com.kramerius.webclient.fedoraws.ComparisonOperator;
import com.kramerius.webclient.fedoraws.FedoraAPIA;
import com.kramerius.webclient.fedoraws.FedoraAPIABindingSOAPHTTPStub;
import com.kramerius.webclient.fedoraws.FedoraAPIAServiceLocator;
import com.kramerius.webclient.fedoraws.FieldSearchQuery;
import com.kramerius.webclient.fedoraws.FieldSearchQueryConditions;
import com.kramerius.webclient.fedoraws.FieldSearchResultResultList;
import com.qbizm.kramerius.webclient.repository.dto.Condition;
import com.qbizm.kramerius.webclient.repository.dto.DatastreamDef;
import com.qbizm.kramerius.webclient.repository.dto.FieldSearchResult;
import com.qbizm.kramerius.webclient.repository.dto.ListSession;
import com.qbizm.kramerius.webclient.repository.dto.MIMETypedStream;
import com.qbizm.kramerius.webclient.repository.dto.ObjectFields;
import com.qbizm.kramerius.webclient.repository.dto.ObjectProfile;
import com.qbizm.kramerius.webclient.repository.dto.Property;
import com.qbizm.kramerius.webclient.repository.dto.RepositoryInfo;

/**
 * Implementation of FEDORA repository interface - WebServices 
 * <p>
 * Copyright (c) 2008 Qbizm technologies, a.s. All rights reserved.
 * <p>
 * This software is the proprietary information of Qbizm technologies, a.s. Use is subject to
 * license terms.
 * <p>
 * Requirements: <br>
 * Keywords: <br>
 * 
 * @author <a href="mailto:pavel.slavicek@qbizm.cz">Pavel Slavicek</a>
 * @version $$ $$
 */
public class RepositoryWSImpl implements Repository {

  /** Logger **/
  private static final Logger LOG = Logger.getLogger(RepositoryWSImpl.class);

  /** URL fragment for Resource Index search **/
  private static final String RI_SEARCH = "/risearch?type=tuples&lang=itql&format=Sparql&query=";

  /** URL fragment for Fedora WebServices **/
  private static final String WS_INTERFACE = "/services/access";

  /** Given Fedora server address **/
  private String serverAddress = null;

  /** Given user name for Fedora server **/
  private String userName = null;

  /** Given user's password for Fedora server **/
  private String password = null;

  /** API of Fedora WebServices **/
  private FedoraAPIA repositoryApi = null;

  /**
   * Hidden constructor
   */
  @SuppressWarnings("unused")
  private RepositoryWSImpl() {
    super();
  }

  /**
   * Constructor
   * @param userName for Fedora server
   * @param password for given user name
   * @param serverAddress URL of Fedora server
   */
  public RepositoryWSImpl(String userName, String password, String serverAddress) {
    this.serverAddress = serverAddress;
    this.userName = userName;
    this.password = password;
    if (serverAddress!=null && serverAddress.length()>0 && serverAddress.endsWith("/")) {
      serverAddress = serverAddress.substring(0, serverAddress.length() - 1);
    }
    try {
      FedoraAPIAServiceLocator serviceLocator = new FedoraAPIAServiceLocator();
      serviceLocator.setFedoraAPIAPortSOAPHTTPEndpointAddress(serverAddress + WS_INTERFACE);
      repositoryApi = (FedoraAPIABindingSOAPHTTPStub) serviceLocator.getFedoraAPIAPortSOAPHTTP();
      ((FedoraAPIABindingSOAPHTTPStub) repositoryApi).setUsername(userName);
      ((FedoraAPIABindingSOAPHTTPStub) repositoryApi).setPassword(password);
    } catch (Exception e) {
      LOG.error("Cannot create instance: " + e.getMessage(), e);
    }
  }

  /**
   * @see {@link Repository#getRepositoryInfo()} 
   */
  public RepositoryInfo getRepositoryInfo() throws RepositoryException {
    if (this.repositoryApi == null) {
      throw new RepositoryException("Connection to FEDORA repository is not established!");
    }
    RepositoryInfo repositoryInfo = null;
    try {
      com.kramerius.webclient.fedoraws.RepositoryInfo repInfoWs = repositoryApi.describeRepository();
      if (repInfoWs != null) {
        // Populate RepositoryInfo object from WS layer into object from this package
        repositoryInfo = new RepositoryInfo();
        repositoryInfo.setDefaultExportFormat(repInfoWs.getDefaultExportFormat());
        repositoryInfo.setOAINamespace(repInfoWs.getOAINamespace());
        repositoryInfo.setRepositoryBaseURL(repInfoWs.getRepositoryBaseURL());
        repositoryInfo.setRepositoryName(repInfoWs.getRepositoryName());
        repositoryInfo.setRepositoryPIDNamespace(repInfoWs.getRepositoryPIDNamespace());
        repositoryInfo.setRepositoryVersion(repInfoWs.getRepositoryVersion());
        repositoryInfo.setSampleAccessURL(repInfoWs.getSampleAccessURL());
        repositoryInfo.setSampleOAIIdentifier(repInfoWs.getSampleOAIIdentifier());
        repositoryInfo.setSampleOAIURL(repInfoWs.getSampleOAIURL());
        repositoryInfo.setSamplePID(repInfoWs.getSamplePID());
        repositoryInfo.setSampleSearchURL(repInfoWs.getSampleSearchURL());
        // Transforms array object into lists
        if (repInfoWs.getAdminEmailList() != null) {
          List<String> admins = new ArrayList<String>();
          for (String email : repInfoWs.getAdminEmailList().getItem()) {
            admins.add(email);
          }
          repositoryInfo.setAdminEmailList(admins);
        }
        if (repInfoWs.getRetainPIDs() != null) {
          List<String> pids = new ArrayList<String>();
          for (String pid : repInfoWs.getRetainPIDs().getItem()) {
            pids.add(pid);
          }
          repositoryInfo.setRetainPIDs(pids);
        }
      } else {
        throw new RemoteException(
          "Fedora WS API returned null com.kramerius.webclient.fedoraws.RepositoryInfo object");
      }
    } catch (RemoteException e) {
      LOG.error("Cannot read repository info: " + e.getMessage(), e);
    }
    return repositoryInfo;
  }
  
  /**
   * @see {@link Repository#getObjectProfile(String, String)
   */
  public ObjectProfile getObjectProfile(String pid, java.lang.String asOfDateTime)  throws RepositoryException {
    if (this.repositoryApi == null) {
      throw new RepositoryException("Connection to FEDORA repository is not established!");
    }
    ObjectProfile profile = null;
    try {
      com.kramerius.webclient.fedoraws.ObjectProfile profWs = repositoryApi.getObjectProfile(pid, asOfDateTime);
      if (profWs!=null) {
        profile = new ObjectProfile();
        if (profWs.getObjModels()!=null && profWs.getObjModels().getModel()!=null) {
          List<String> models = new LinkedList<String>();
          for (String model : profWs.getObjModels().getModel()) {
             models.add(model);
          }
          profile.setObjModels(models);
        }
        profile.setObjCreateDate(profWs.getObjCreateDate());
        profile.setObjDissIndexViewURL(profWs.getObjDissIndexViewURL());
        profile.setObjItemIndexViewURL(profWs.getObjItemIndexViewURL());
        profile.setObjLabel(profWs.getObjLabel());
        profile.setObjLastModDate(profWs.getObjLastModDate());
        profile.setPid(profWs.getPid());
      }
    } catch (Exception e) {
      LOG.error("Cannot read object profile: " + e.getMessage(), e);
      throw new RepositoryException("Cannot get object profile from repository");
    }
    return profile;
  }
  
  /**
   * @see {@link Repository#getDataStreams(String, String)
   */
  public List<DatastreamDef> getDataStreamsInfo(String pid, String asOfDateTime) throws RepositoryException {
    if (this.repositoryApi == null) {
      throw new RepositoryException("Connection to FEDORA repository is not established!");
    }
    if (pid == null || pid.trim().length()==0) {
      throw new RepositoryException("You cannot call this method with empty PID!");
    }        
    List<DatastreamDef> streams = new LinkedList<DatastreamDef>();
    try {
      com.kramerius.webclient.fedoraws.DatastreamDef[] dsWs = repositoryApi.listDatastreams(pid, asOfDateTime);
      if (dsWs!=null) {
        for (com.kramerius.webclient.fedoraws.DatastreamDef streamWs : dsWs) {
          DatastreamDef stream = new DatastreamDef();
          stream.setID(streamWs.getID());
          stream.setLabel(streamWs.getLabel());
          stream.setMIMEType(streamWs.getMIMEType());
          streams.add(stream);
        }
      }
    } catch (Exception e) {
      LOG.error("Cannot read object's datastreams information: " + e.getMessage(), e);
      throw new RepositoryException(e);
    }
    return streams;
  }
  
  /**
   * @See {@link Repository#getDatastream(String, String, String)}
   */
  public MIMETypedStream getDatastream(String pid, String dsID, String asOfDateTime) throws RepositoryException {
    if (this.repositoryApi == null) {
      throw new RepositoryException("Connection to FEDORA repository is not established!");
    }
    if (pid == null || pid.trim().length()==0) {
      throw new RepositoryException("You cannot call this method with empty PID!");
    }        
    if (dsID == null || dsID.trim().length()==0) {
      throw new RepositoryException("You cannot call this method with empty dsID!");
    }            
    MIMETypedStream stream = null;
    try {
      com.kramerius.webclient.fedoraws.MIMETypedStream mimeWs = repositoryApi.getDatastreamDissemination(pid, dsID, asOfDateTime);
      if (mimeWs!=null) {
        stream = new MIMETypedStream();
        stream.setMIMEType(mimeWs.getMIMEType());
        stream.setStream(mimeWs.getStream());
        
        List<Property> properties = new ArrayList<Property>();
        if (mimeWs.getHeader()!=null && mimeWs.getHeader().getProperty()!=null) {
          for (com.kramerius.webclient.fedoraws.Property propWs : mimeWs.getHeader().getProperty()) {
            Property prop = new Property();
            prop.setName(propWs.getName());
            prop.setValue(propWs.getValue());
            properties.add(prop);
          }
        }
        stream.setProperties(properties);
      }
    } catch (Exception e) {
      LOG.error("Cannot read object's datastream: " + e.getMessage(), e);
      throw new RepositoryException("Cannot read object's datastream:" + e.getMessage());
    }
    return stream;
  }

  /**
   * @see {@link Repository#findObjects(List, BigInteger, String, List)} 
   */
  public FieldSearchResult findObjects(
    List<String> resultFields,
    BigInteger maxResults,
    String searchTerms,
    List<Condition> conditions) throws RepositoryException {

    if (this.repositoryApi == null) {
      throw new RepositoryException("Connection to FEDORA repository is not established!");
    }
    if (resultFields == null || resultFields.isEmpty()) {
      throw new RepositoryException("You must specify result fields!");
    }
    if (maxResults == null || maxResults.intValue() < 0) {
      throw new RepositoryException("Maxresult must be not negative number!");
    }
    FieldSearchResult results = null;
    try {
      // Transforms result fields into array
      String resFiledsArray[] = new String[resultFields.size()];
      for (int i = 0; i < resFiledsArray.length; i++) {
        resFiledsArray[i] = resultFields.get(i);
      }
      ArrayOfString resFields = new ArrayOfString(resFiledsArray);
      FieldSearchQuery query = new FieldSearchQuery();
      query.setTerms(searchTerms);
      // Transforms Conditions for search query
      if (conditions != null && !conditions.isEmpty()) {
        FieldSearchQueryConditions fconds = new FieldSearchQueryConditions();
        List<com.kramerius.webclient.fedoraws.Condition> conds = new ArrayList<com.kramerius.webclient.fedoraws.Condition>();
        for (Condition con : conditions) {
          com.kramerius.webclient.fedoraws.Condition wsCond = new com.kramerius.webclient.fedoraws.Condition();
          wsCond.setProperty(con.getProperty());
          wsCond.setValue(con.getValue());
          wsCond.setOperator(ComparisonOperator.fromString(con.getOperator().getValue()));
          conds.add(wsCond);
        }
        if (!conds.isEmpty()) {
          com.kramerius.webclient.fedoraws.Condition condWs[] = new com.kramerius.webclient.fedoraws.Condition[conds.size()];
          for (int i = 0; i < condWs.length; i++) {
            condWs[i] = conds.get(i);
          }
          fconds.setCondition(condWs);
        }
        query.setConditions(fconds);
      }
      // Calls Fedora web service
      com.kramerius.webclient.fedoraws.FieldSearchResult res = repositoryApi.findObjects(
        resFields,
        new NonNegativeInteger(maxResults.toByteArray()),
        query);
      // Transforms results
      if (res != null) {
        results = new FieldSearchResult();
        if (res.getListSession() != null) {
          ListSession session = new ListSession();
          session.setCompleteListSize(res.getListSession().getCompleteListSize());
          if (res.getListSession().getCursor() != null) {
            session.setCursor(new BigInteger(res.getListSession().getCursor().toByteArray()));
          }
          session.setExpirationDate(res.getListSession().getExpirationDate());
          session.setToken(res.getListSession().getToken());
          results.setListSession(session);
        }
        FieldSearchResultResultList resList = res.getResultList();
        if (resList != null) {
          com.kramerius.webclient.fedoraws.ObjectFields[] objectFiledsWs = resList.getObjectFields();
          if (objectFiledsWs != null) {
            List<ObjectFields> resultList = new ArrayList<ObjectFields>(objectFiledsWs.length);
            for (com.kramerius.webclient.fedoraws.ObjectFields objWs : objectFiledsWs) {
              resultList.add(populateObjectFields(objWs));
            }
            results.setResultList(resultList);
          }
        }
      }
    } catch (Exception e) {
      LOG.error("Cannot findObjects: " + e.getMessage(), e);
    }
    return results;
  }

  /**
   * @see {@link Repository#riSearch(String, Class)
   */
  public List<Object> riSearch(String query, Class<?> targetObject) throws RepositoryException {
    List<Object> tmp = null;
    String riSearchInterface = this.serverAddress + RI_SEARCH;
    try {
      riSearchInterface += URLEncoder.encode(query, "UTF-8");
      URLConnection connection = (new URL(riSearchInterface)).openConnection();
      if (this.userName != null) {
        connection.setRequestProperty("Authorization", "Basic "
          + org.apache.axis.encoding.Base64.encode((userName + ":" + password).getBytes()));
      }
      InputStream is = connection.getInputStream();
      int read;
      byte[] b = new byte[1024];
      StringBuffer sb = new StringBuffer();
      while ((read = is.read(b, 0, 1024)) >= 0) {
        sb.append(new String(b, 0, read));
      }
      SparqlParser parser = new SparqlParser(targetObject, sb.toString().getBytes());
      tmp = parser.getResult();
    } catch (Exception e) {
      LOG.error("riSearch query: " + query);
      LOG.error("riSearch Class<?>: " + targetObject);
      LOG.error("riSearch: " + e.getMessage(), e);
      throw new RepositoryException(e.getMessage());
    }
    return tmp;
  }
  
  /**
   * Populates ObjectField object from WS layer into ObjectField
   * from this package
   * @param objFieldsWs
   * @return
   */
  private static ObjectFields populateObjectFields(
    com.kramerius.webclient.fedoraws.ObjectFields objFieldsWs) {
    ObjectFields obj = new ObjectFields();
    obj.setPid(objFieldsWs.getPid());
    obj.setLabel(objFieldsWs.getLabel());
    obj.setState(objFieldsWs.getState());
    obj.setOwnerId(objFieldsWs.getOwnerId());
    obj.setCDate(objFieldsWs.getCDate());
    obj.setMDate(objFieldsWs.getMDate());
    obj.setDcmDate(objFieldsWs.getDcmDate());
    obj.setTitle(objFieldsWs.getTitle());
    obj.setCreator(objFieldsWs.getCreator());
    obj.setSubject(objFieldsWs.getSubject());
    obj.setDescription(objFieldsWs.getDescription());
    obj.setPublisher(objFieldsWs.getPublisher());
    obj.setContributor(objFieldsWs.getContributor());
    obj.setDate(objFieldsWs.getDate());
    obj.setType(objFieldsWs.getType());
    obj.setFormat(objFieldsWs.getFormat());
    obj.setIdentifier(objFieldsWs.getIdentifier());
    obj.setSource(objFieldsWs.getSource());
    obj.setLanguage(objFieldsWs.getLanguage());
    obj.setRelation(objFieldsWs.getRelation());
    obj.setCoverage(objFieldsWs.getCoverage());
    obj.setRights(objFieldsWs.getRights());    
    return obj;
  }
  
  /**
   * @see {@link Repository#getObjectFileds(String, String)
   */
  public ObjectFields getObjectFileds(String fields, String pid) throws RepositoryException {
    if (pid==null || pid.trim().length()==0) {
      throw new RepositoryException("Cannot find digital object with null pid");
    }
    ObjectFields tmp = null;
    try {
      
      List<String> searchFieldsList = new LinkedList<String>();
      if (fields!=null) {
        StringTokenizer st = new StringTokenizer(fields, ",");
        while (st.hasMoreTokens()) {
          String el = st.nextToken();
          if (el!=null) {
            searchFieldsList.add(el.trim());
          }
        }
      }
      if (!searchFieldsList.contains("pid")) {
        searchFieldsList.add("pid");              
      }
      List<Condition> conditions = new LinkedList<Condition>();
      conditions.add(new Condition("pid", com.qbizm.kramerius.webclient.repository.dto.ComparisonOperator.eq, pid));
      FieldSearchResult result = this.findObjects(searchFieldsList, new BigInteger("1"), null, conditions);
      if (result!=null && result.getResultList()!=null && !result.getResultList().isEmpty()) {
        tmp = result.getResultList().get(0);
      }      
    } catch (Exception e) {
      LOG.error("Cannot get object's fields : " + e.getMessage(), e);
      throw new RepositoryException(e.getMessage());
    }
    return tmp;
  }
}
