package com.qbizm.kramerius.webclient;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;

import com.qbizm.kramerius.webclient.repository.Repository;
import com.qbizm.kramerius.webclient.repository.RepositoryException;
import com.qbizm.kramerius.webclient.repository.RepositoryFactory;
import com.qbizm.kramerius.webclient.repository.dto.ComparisonOperator;
import com.qbizm.kramerius.webclient.repository.dto.Condition;
import com.qbizm.kramerius.webclient.repository.dto.FieldSearchResult;
import com.qbizm.kramerius.webclient.repository.dto.RepositoryInfo;

/**
 * Repository bean implementation, provides basic
 * operations with FEDORA communication layer 
 * <p>
 * Copyright (c) 2008 Qbizm technologies, a.s. All rights reserved.
 * <p>
 * This software is the proprietary information of Qbizm technologies, a.s. Use is subject to
 * license terms.
 * <p>
 * Requirements: <br>
 * Keywords: <br>
 * 
 * @author <a href="mailto:pavel.slavicek@qbizm.cz">Pavel Slavicek</a>
 * @version $$ $$
 */
public class RepositoryBean {
  
  /** Logger **/
  private static final Logger LOG = Logger.getLogger(RepositoryBean.class);
  
  /** Configuration names **/
  private static final String CONFIGURATION_FILE = "configuration.properties";
  private static final String CONFVAL_URL = "repository.url";
  private static final String CONFVAL_USERNAME = "repository.username";
  private static final String CONFVAL_PASSWORD = "repository.password";
  private static final String CONFVAL_IMPL = "repository.implementation";
  
  /** Digital objects definition configurations **/
  private static final String CONFVAL_DIGIOBJECT_ID_MAKRO   = "[ID]";
  private static final String CONFVAL_DIGIOBJECT_PREFIX     = "repository.digital.object.";
  private static final String CONFVAL_DIGIOBJECT_PID        = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO;
  private static final String CONFVAL_DIGIOBJECT_ROOT       = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO +".root";
  private static final String CONFVAL_DIGIOBJECT_ITQLITEMS  = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO +".itqlitems";
  private static final String CONFVAL_DIGIOBJECT_SUBOBJECT  = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO +".subobject";
  private static final String CONFVAL_DIGIOBJECT_ITQLPARENT = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO +".itqlparent";
  private static final String CONFVAL_DIGIOBJECT_STYLESHEET = CONFVAL_DIGIOBJECT_PREFIX + CONFVAL_DIGIOBJECT_ID_MAKRO +".stylesheet";
  
  /** Repository access point **/
  private Repository repository = null;
  
  /** Repository configuration values **/
  private String repositoryUrl = null;
  private String repositoryUserName = null;
  private String repositoryPassword = null;
  private String repositoryImplementation = null;

  /**
   * Digital objects representations
   */
  private Map<String, DigitalObjectTypeDef> digitalObjectDefs = new HashMap<String, DigitalObjectTypeDef>();
  
  private Map<Integer, String> configIdToPidMapping = new HashMap<Integer, String>();
  
  /** 
   * Types of configuration value  
   */
  private enum DigiObjConfType {
    PID, ROOT, ITQLITEMS, SUBOBJECT, ITQLPARENT, STYLESHEET, UNKNOWN;
  }

  
  /**
   * Constructor
   * Reads repository configuration
   */
  public RepositoryBean() {
    super();
    LOG.info("Reading configuration");    
    Properties properties = new Properties();
    try {
      // Reads configuration values from CONFIGURATION_FILE file
      properties.load(this.getClass().getClassLoader().getResourceAsStream(CONFIGURATION_FILE));
      this.repositoryUrl = properties.getProperty(CONFVAL_URL);
      this.repositoryUserName  = properties.getProperty(CONFVAL_USERNAME);
      this.repositoryPassword  = properties.getProperty(CONFVAL_PASSWORD);
      this.repositoryImplementation  = properties.getProperty(CONFVAL_IMPL);
      this.repositoryUrl = (this.repositoryUrl!=null) ? this.repositoryUrl.trim() : null;
      this.repositoryUserName = (this.repositoryUserName!=null) ? this.repositoryUserName.trim() : null;
      this.repositoryPassword = (this.repositoryPassword!=null) ? this.repositoryPassword.trim() : null;
      this.repositoryImplementation = (this.repositoryImplementation!=null) ? this.repositoryImplementation.trim() : null;      

      // Stupid solution, but no time no fun
      for (Object key : properties.keySet()) {
        if (key instanceof String) {
          String strKey = (String) key;
          String value = properties.getProperty(strKey);
          value = (value!=null) ? value.trim() : null;
          if (isDigitalObjectConfValue(strKey)) {
            Integer idConf = getIdOfDigitalObjectFromConfVal(strKey);
            DigiObjConfType confType = getPartTypeOfDigiObjConf(strKey, idConf);
            if (confType==DigiObjConfType.PID) {
              processDigObjConfiguration(null, value, confType, idConf);
            } else {
              LOG.warn("Cannot read digital object configuration from property file : [" + strKey + "," + value + "]");
            }
          }
        }
      }
      
      for (Object key : properties.keySet()) {
        if (key instanceof String) {
          String strKey = (String) key;
          String value = properties.getProperty(strKey);
          if (isDigitalObjectConfValue(strKey)) {
            Integer idConf = getIdOfDigitalObjectFromConfVal(strKey);
            DigiObjConfType confType = getPartTypeOfDigiObjConf(strKey, idConf);
            if (confType!=DigiObjConfType.UNKNOWN && confType!=DigiObjConfType.PID) {
              String pid = null;
              if (this.configIdToPidMapping.containsKey(idConf)) {
                pid = this.configIdToPidMapping.get(idConf);  
              }              
              processDigObjConfiguration(pid, value, confType, idConf);
            } else {
              LOG.warn("Cannot read digital object configuration from property file : [" + strKey + "," + value + "]");
            }
          }
        }
      }
    } catch (Exception e) {
      LOG.error("Cannot read repository configuration (" + CONFIGURATION_FILE + ") :" + e.getMessage(), e);
    }
    LOG.info("Using URL for access Fedora repository : " + repositoryUrl);
    LOG.info("Repository user's name : " + repositoryUserName);
    LOG.info("Repository implementation : " + repositoryImplementation);
    repository = RepositoryFactory.getInstance(repositoryUserName, repositoryPassword, repositoryUrl, repositoryImplementation);
    if (repository!=null) {
      LOG.info("Connected to repository.");
      try {
        RepositoryInfo info = repository.getRepositoryInfo();
        if (info!=null) {
          LOG.info("Repository version : " + info.getRepositoryVersion());
          LOG.info("Repository base URL : " + info.getRepositoryBaseURL());
          LOG.info("Repository name : " + info.getRepositoryName());
          LOG.info("Administrators emails : " + info.getAdminEmailList());
        }
      } catch(RepositoryException e) {
        LOG.error("Cannot get repository info: " + e.getMessage(), e);
      }
      if (!this.digitalObjectDefs.isEmpty()) {
        LOG.info("Loaded digital object's configurations");
        for (String pid : this.digitalObjectDefs.keySet()) {
          LOG.info(this.digitalObjectDefs.get(pid));
        }
        LOG.info("End of loaded digital object's configurations");
      }
      LOG.info("Loading titles for root digital objects");
      List<DigitalObjectTypeDef> rootItems = getRootDigitalObjectDefs();
      if (rootItems!=null && !rootItems.isEmpty()) {
        List<String> fields = new LinkedList<String>();
        fields.add("title");
        fields.add("pid");
        for (DigitalObjectTypeDef item : rootItems) {
          LOG.info("Loading title for model : " + item.getPid());
          try {
            List<Condition> conditions = new LinkedList<Condition>();
            conditions.add(new Condition("pid", ComparisonOperator.eq, item.getPid()));
            FieldSearchResult result = this.repository.findObjects(fields, new BigInteger("1"), null, conditions);
            if (result!=null && result.getResultList()!=null && !result.getResultList().isEmpty()) {
              item.setTitle(result.getResultList().get(0).getTitlesView());
            }
            LOG.info("Loaded title for model : " + item.getTitle());
          } catch (Exception e) {
            LOG.error("Cannot load title for root digital object " + item.getPid());
          }
        }
      }
    }
    LOG.info("End of creating connection to repository");
  }
  
  /**
   * Returns current instance of digital repository object
   * @return
   */
  public Repository getRepository() {
    return this.repository;
  }
  
  /**
   * Returns map of digital object definitions
   * @return
   */
  public Map<String, DigitalObjectTypeDef> getDigitalObjectDefs() {
    return digitalObjectDefs;
  }
  
  /**
   * Sets digital object definitions
   * @param digitalObjectDefs
   */
  public void setDigitalObjectDefs(Map<String, DigitalObjectTypeDef> digitalObjectDefs) {
    this.digitalObjectDefs = digitalObjectDefs;
  }
  
  /**
   * Returns definition of digital object with given PID.
   * If PID starts with prefix info:fedora/ this prefix will be
   * removed.
   * @param pid
   * @return
   */
  public DigitalObjectTypeDef getDigitalObjectDef(String pid) {
    if (pid!=null && pid.startsWith("info:fedora/")) {
      pid = pid.substring("info:fedora/".length());
    }
    return this.digitalObjectDefs.get(pid);
  }
  
  /**
   * Returns list of root digital objects
   * @return
   */
  public List<DigitalObjectTypeDef> getRootDigitalObjectDefs() {
    List<DigitalObjectTypeDef> tmp = new LinkedList<DigitalObjectTypeDef>();
    if (this.digitalObjectDefs!=null) {
      for (String pid : this.digitalObjectDefs.keySet()) {
        DigitalObjectTypeDef def = getDigitalObjectDef(pid);
        if (def.isRoot()) {
          tmp.add(def);
        }
      }
    }
    return tmp;
  }
  
  /**
   * Returns URL of repository from configuration
   */
  public String getRepositoryUrl() {
    return this.repositoryUrl;
  }

  /**
   * Returns true if the given String value is digital
   * object definition
   * @param value
   * @return
   */
  private static boolean isDigitalObjectConfValue(String value) {
    return (value!=null && value.startsWith(CONFVAL_DIGIOBJECT_PREFIX));    
  }

  /**
   * Reads id of digital object configuration
   * @param confName
   * @return
   */
  private static Integer getIdOfDigitalObjectFromConfVal(String confName) {
    Integer id = null;
    if (confName!=null && confName.startsWith(CONFVAL_DIGIOBJECT_PREFIX)) {
      String tmp = confName.substring(CONFVAL_DIGIOBJECT_PREFIX.length());
      if (tmp!=null) {
        int index = tmp.indexOf(".");
        if (index>0) {
          tmp = tmp.substring(0, index);
        }
      }
      try {
        id = Integer.valueOf(tmp);
      } catch (NumberFormatException e) {
        // Ignore exception
      }
    }
    return id;
  }

  /**
   * Returns type of digital's object configuration from configuration name
   * @param confValue
   * @param idOfConf
   * @return
   */
  private static DigiObjConfType getPartTypeOfDigiObjConf(String confValue, Integer idOfConf) {
    DigiObjConfType type = DigiObjConfType.UNKNOWN;
    if (confValue!=null && idOfConf!=null) {
      String pid = CONFVAL_DIGIOBJECT_PID.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));
      String root = CONFVAL_DIGIOBJECT_ROOT.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));
      String itqlitems = CONFVAL_DIGIOBJECT_ITQLITEMS.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));
      String childs = CONFVAL_DIGIOBJECT_SUBOBJECT.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));
      String itqlparent = CONFVAL_DIGIOBJECT_ITQLPARENT.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));
      String stylesheet = CONFVAL_DIGIOBJECT_STYLESHEET.replace(CONFVAL_DIGIOBJECT_ID_MAKRO, Integer.toString(idOfConf));      
      if (pid.equals(confValue)) {
        type = DigiObjConfType.PID;
      } else if (root.equals(confValue)) {
        type = DigiObjConfType.ROOT;
      } else if (itqlitems.equals(confValue)) {
        type = DigiObjConfType.ITQLITEMS;
      } else if (confValue.startsWith(childs)) {
        type = DigiObjConfType.SUBOBJECT;        
      } else if (itqlparent.equals(confValue)) {
        type = DigiObjConfType.ITQLPARENT;
      } else if (confValue.startsWith(stylesheet)) {
        type = DigiObjConfType.STYLESHEET;        
      } else {
        type = DigiObjConfType.UNKNOWN;
      }
    }
    return type;
  }

  /**
   * Process digital object's configuration value 
   * @param pid
   * @param value
   * @param type
   */
  private void processDigObjConfiguration(String pid, String value, DigiObjConfType type, Integer confId) {
    if (type==DigiObjConfType.PID) {
      this.digitalObjectDefs.put(value, new DigitalObjectTypeDef(value, value, false, null, null, null));
      this.configIdToPidMapping.put(confId, value);
    } else if (type==DigiObjConfType.ROOT) {
      /** Is Root digital object? **/
      boolean root = false;
      try {
        root = Boolean.valueOf(value);
      } catch (Exception e) {
        LOG.warn("Cannot interpret value " + value + " as boolean value");
      }
      if (this.digitalObjectDefs.containsKey(pid)) {
        this.digitalObjectDefs.get(pid).setRoot(root);
      } else {
        LOG.warn("Cannot find digital object " + pid);
      }      
    } else if (type==DigiObjConfType.ITQLITEMS) {
      /** ITQL for items **/
      if (this.digitalObjectDefs.containsKey(pid)) {
        this.digitalObjectDefs.get(pid).setSubItems(value);
      } else {
        LOG.warn("Cannot find digital object " + pid);
      }
    } else if (type==DigiObjConfType.SUBOBJECT) {
      /** SubObjects **/
      DigitalObjectTypeDef.SubItemDef subItemDef = null;
      if (value!=null) {
        String subObjectDef[] = value.split("\\|");        
        if (subObjectDef.length>1) {
          subItemDef = new DigitalObjectTypeDef.SubItemDef();
          subItemDef.setTitle(subObjectDef[0]);
        }
        if (subObjectDef.length>=2) {
          subItemDef.setItql(subObjectDef[1]);
        }
        if (subObjectDef.length>=3) {
          subItemDef.addDataStreams(subObjectDef[2]);
        }
      }
      if (this.digitalObjectDefs.containsKey(pid) && subItemDef!=null) {
        this.digitalObjectDefs.get(pid).getSubItems().add(subItemDef);
      } else {
        LOG.warn("Cannot find digital object " + pid);
      }
    } else if (type==DigiObjConfType.ITQLPARENT) {
      /** ITQL for parent **/
      if (this.digitalObjectDefs.containsKey(pid)) {
        this.digitalObjectDefs.get(pid).setItqlParent(value);
      } else {
        LOG.warn("Cannot find digital object " + pid);
      }
    } else if (type==DigiObjConfType.STYLESHEET) {
      /** StyleSheets for XSL transformation **/
      if (this.digitalObjectDefs.containsKey(pid)) {
        if (value!=null) {
          String dcDef[] = value.split("\\|");
          if (dcDef!=null && dcDef.length==2) {
            this.digitalObjectDefs.get(pid).addStylesheet(dcDef[0], dcDef[1]);            
          } else {
            LOG.warn("Cannot parse information about stylesheets for datastream from value " + value);
          }          
        }        
      } else {
        LOG.warn("Cannot find digital object " + pid);
      }
    }
  }
}


  